from sklearn.ensemble import RandomForestClassifier, AdaBoostClassifier
from sklearn.svm import SVC
from sklearn.neighbors import KNeighborsClassifier
from sklearn.linear_model import LogisticRegression
from xgboost import XGBClassifier
from catboost import CatBoostClassifier
import sklearn
from sklearn.metrics import *
from category_encoders import TargetEncoder
from imblearn.under_sampling import RandomUnderSampler
from imblearn.over_sampling import SMOTE
from onepiecepredictor.OnePiecePredictor2 import OnePiecePredictor3
from sklearn.model_selection import train_test_split

class OnePieceClassifier(OnePiecePredictor3):
    """
    For hyper parameter tuning with cross validation and stratified splitting of data if required.

    X -> array-like(supported by Sklearn). If testTrainSplit is passed, this will be split into train and test
    Y -> array-like(supported by Sklearn). If testTrainSplit is passed, this will be split into train and test
    model -> string Currently supported models: LOGISTIC,RF,SVM,KNN,ADABOOST,XGBOOST,CATBOOST
    testX -> array-like(supported by Sklearn), test data. Ignored if testTrainSplit is passed
    testY -> array-like(supported by Sklearn), test data. Ignored if testTrainSplit is passed
    testTrainSplit -> float, ratio passed will be the amount of test data.
    stratify -> bool, used to perform stratified splitting. If passed data will be split based on Y.
    hyperParams -> dictionary, Hyper parameters specific to the model passed. If passed CV is performed.
    performCV -> bool, Used when hyperParams not passed to perform plain CV.
    folds -> int, No of folds to be used for CV.
    applySmote -> bool, To apply smote to oversample the data. Pass only one of applySmote or underSample
    underSample -> bool, To randomly undersample the majority data.
    sampling -> str, Values supported by SMOTE, RandomUnderSampler classes in imblearn library.
    scoring -> str, Evaluation metric. Currently supported values: accuracy,balanced_accuracy,f1,precision,recall,roc_auc. If not passed accuracy is used.
    targetEncodeCols -> List. List of columns to target encode.
    modelParams -> dictionary, Any model specific parameters can be passed as dictionary.
    multiClass -> Pass true in case of multiclass classification.
    """
    def __init__(self, X, Y, model ,testX = None, testY = None,testTrainSplit = None,
                stratify = None, hyperParams = None, performCV = None, folds = 5,
                applySmote = False, underSample = False, sampling = None,
                scoring = None,  targetEncodeCols = None, multiClass = False, modelParams = None):
        if(modelParams is None):
            modelParams = {}
        self.multiClass = multiClass
        self.applySmote = applySmote
        self.sampling = sampling
        self.stratify = stratify
        self.underSample = underSample
        super().__init__(
            X = X, Y = Y, testX=testX, testY=testY, testTrainSplit=testTrainSplit,
            model=model, folds=folds, hyperParams=hyperParams, scoring=scoring, performCV=performCV,
            targetEncodeCols = targetEncodeCols,
            modelParams = modelParams)
        if(not self.scoring):
            self.scoring = 'accuracy'
        self.scoreToFuncDict = self.__getScoreToFuncDict()

    def getTestTrainSlipt(self):
        ## If both testX and testTrainSplit are not passed throw exception.
        if ((self.testX is None) and (self.testTrainSplit is None)):
            raise Exception("Please pass testX or testTrainSplit")

        ## If targetEncodeCols is given first target encode them.
        if (self.targetEncodeCols):
            for col in self.targetEncodeCols:
                encoder = TargetEncoder()
                self.X[col] = encoder.fit_transform(self.X[col], self.Y)
                if(self.testX and self.testY):
                    self.testX[col] = encoder.fit_transform(self.testX[col], self.testY)

        if ((self.testX is not None) and (self.testTrainSplit is None)):
            return self.X, self.testX, self.Y, self.testY

        ## If stratify, smote and testTrainSplits are not passed, then just return.
        if (not self.stratify and not self.applySmote and not self.testTrainSplit):
            return self.X, self.testX, self.Y, self.testY

        # If startify flag is passed then stratify it using Y variable.
        startifyVar = self.Y if self.stratify else None
        X_train, X_test, y_train, y_test = train_test_split(self.X, self.Y, stratify=startifyVar,
                                                            test_size=self.testTrainSplit, random_state = 7)
        if (not self.applySmote and not self.underSample):
            return X_train, X_test, y_train, y_test
        else:
            X_train = X_train if self.testTrainSplit is not None else self.X
            y_train = y_train if self.testTrainSplit is not None else self.Y
            X_test = X_test if self.testTrainSplit is not None else self.testX
            y_test = y_test if self.testTrainSplit is not None else self.testY
            if (self.applySmote):
                sm = SMOTE(sampling_strategy=self.sampling)
                X_train_res, y_train_res = sm.fit_sample(X_train, y_train)
                return X_train_res, X_test, y_train_res, y_test
            if (self.underSample):
                underSampler = RandomUnderSampler(sampling_strategy=self.sampling)
                X_train_res, y_train_res = underSampler.fit_sample(X_train, y_train)
                return X_train_res, X_test, y_train_res, y_test

    def fit(self):
        super().fit()

    def test(self):
        super().test()

    def predict(self):
        """
        Returns score and predictions.
        """
        if(self.testY is None):
            return 0, self.test()

        preds = self.bestEstimator.predict(self.testX)
        res = getattr(sklearn.metrics, self.scoreToFuncDict[self.scoring])(self.testY, preds)
        print(self.model, self.scoring, res)
        return res, preds

    def newDataPredict(self, testData):
        super().newDataPredict(testData)

    def getEstimatorModel(self):

        specialMultiClassClassifiers = ["LOGISTIC"]
        modelToClassMapper = {
            "LOGISTIC" : LogisticRegression(),
            "RF" : RandomForestClassifier(),
            "SVM" : SVC(),
            "KNN" : KNeighborsClassifier(),
            "ADABOOST" : AdaBoostClassifier(),
            "XGBOOST" : XGBClassifier(),
            "CATBOOST" : CatBoostClassifier(),
        }

        if(self.model not in modelToClassMapper):
            raise Exception("Please pass a valid model from: LOGISTIC,RF,SVM,KNN,ADABOOST,XGBOOST,CATBOOST")
        else:
            cls = modelToClassMapper[self.model]
            if ((self.multiClass) and (self.model in specialMultiClassClassifiers) and (('multi_class' not in self.modelParams))):
                self.modelParams['multi_class'] = 'ovr'

            if ((self.model == "CATBOOST") and ('logging_level' not in self.modelParams)):
                self.modelParams['logging_level'] = 'Silent'

            return cls.set_params(**self.modelParams)

    def __getScoreToFuncDict(self):

        scoreToFuncDict = {
            "accuracy" : 'accuracy_score',
            "balanced_accuracy" : 'balanced_accuracy_score',
            "f1" : 'f1_score',
            "precision": 'precision_score',
            "recall" : 'recall_score',
            "roc_auc" : 'roc_auc_score'
        }

        if(self.scoring and self.scoring not in scoreToFuncDict):
            raise Exception("Please pass a valid scoring from: accuracy,balanced_accuracy,f1,precision,recall,roc_auc")

        return scoreToFuncDict




from sklearn.ensemble import RandomForestRegressor, AdaBoostRegressor
from sklearn.svm import SVR
from sklearn.neighbors import KNeighborsRegressor
from sklearn.linear_model import LinearRegression
from xgboost import XGBRegressor
from catboost import CatBoostRegressor
import sklearn
from sklearn.metrics import *
from category_encoders import TargetEncoder
from sklearn.model_selection import train_test_split
from onepiecepredictor.OnePiecePredictor2 import OnePiecePredictor3

class OnePieceRegression(OnePiecePredictor3):
    """
        For hyper parameter tuning with cross validation and stratified splitting of data if required.

        X -> array-like(supported by Sklearn). If testTrainSplit is passed, this will be split into train and test
        Y -> array-like(supported by Sklearn). If testTrainSplit is passed, this will be split into train and test
        model -> string Currently supported models: LOGISTIC,RF,SVM,KNN,ADABOOST,XGBOOST,CATBOOST
        testX -> array-like(supported by Sklearn), test data. Ignored if testTrainSplit is passed
        testY -> array-like(supported by Sklearn), test data. Ignored if testTrainSplit is passed
        testTrainSplit -> float, ratio passed will be the amount of test data.
        hyperParams -> dictionary, Hyper parameters specific to the model passed. If passed CV is performed.
        performCV -> bool, Used when hyperParams not passed to perform plain CV.
        folds -> int, No of folds to be used for CV.
        scoring -> str, Evaluation metric. Currently supported values: r2,neg_mean_squared_error. If not passed r2 is used.
        targetEncodeCols -> List. List of columns to target encode.
        modelParams -> dictionary, Any model specific parameters can be passed as dictionary.

    """

    def __init__(self, X, Y, model, modelParams = None, testX = None, testY = None,testTrainSplit = None,
                 folds = 5, hyperParams = None,scoring = None, performCV = None, targetEncodeCols = None):
        if (modelParams is None):
            modelParams = {}
        super().__init__(
            X = X, Y = Y, model=model,modelParams = modelParams,testX=testX, testY=testY, testTrainSplit=testTrainSplit,
             folds=folds, hyperParams=hyperParams,scoring=scoring, performCV=performCV, targetEncodeCols = targetEncodeCols,
            )
        self.scoreToFuncDict = self.__getScoreToFuncDict()
        if(not self.scoring):
            self.scoring = 'r2'


    def getTestTrainSlipt(self):
        ## If both testX and testTrainSplit are not passed throw exception.
        if ((self.testX is None) and (self.testTrainSplit is None)):
            raise Exception("Please pass testX or testTrainSplit")

        if (self.targetEncodeCols):
            for col in self.targetEncodeCols:
                encoder = TargetEncoder()
                self.X[col] = encoder.fit_transform(self.X[col])
                if (self.testX):
                    self.testX[col] = encoder.fit_transform(self.testX[col])

        if(self.testTrainSplit):
            X_train, X_test, y_train, y_test = train_test_split(self.X, self.Y, test_size=self.testTrainSplit, random_state=7)
            return X_train, X_test, y_train, y_test
        else:
            return self.X, self.testX, self.Y, self.testY


    def fit(self):
        super().fit()

    def test(self):
        super().test()

    def predict(self):
        """
        Returns score and predictions.
        """
        if (self.testY is None):
            return 0, self.test()

        preds = self.bestEstimator.predict(self.testX)
        res = getattr(sklearn.metrics, self.scoreToFuncDict[self.scoring])(self.testY, preds)
        print(self.model, self.scoring, res)
        return res, preds

    def newDataPredict(self, testData):
        super().newDataPredict(testData)

    def getEstimatorModel(self):

        modelToClassMapper = {
            "LINEAR" : LinearRegression(),
            "RF" : RandomForestRegressor(),
            "SVM" : SVR(),
            "KNN" : KNeighborsRegressor(),
            "ADABOOST" : AdaBoostRegressor(),
            "XGBOOST" : XGBRegressor(),
            "CATBOOST" : CatBoostRegressor(),
        }

        if(self.model not in modelToClassMapper):
            raise Exception("Please pass a valid model from: LINEAR,RF,SVM,KNN,ADABOOST,XGBOOST,CATBOOST")
        else:
            cls = modelToClassMapper[self.model]
            if ((self.model == "CATBOOST") and ('logging_level' not in self.modelParams)):
                self.modelParams['logging_level'] = 'Silent'

            return cls.set_params(**self.modelParams)

    def __getScoreToFuncDict(self):

        scoreToFuncDict = {
            "r2": 'r2_score',
            "neg_mean_squared_error" : 'mean_squared_error'
        }

        return scoreToFuncDict





from sklearn.model_selection import train_test_split
from sklearn.ensemble import RandomForestClassifier
from sklearn.metrics import *
from sklearn.model_selection import GridSearchCV
from category_encoders import TargetEncoder
from imblearn.over_sampling import SMOTE
import gc
from sklearn.model_selection import StratifiedKFold, KFold
from sklearn.model_selection import cross_val_score
import abc
from category_encoders import TargetEncoder
from imblearn.under_sampling import RandomUnderSampler

###################################################################################
## Support Undersampling as well.
## Return Predictions from train as well.
###################################################################################
class OnePiecePredictor3(metaclass=abc.ABCMeta):

    @abc.abstractmethod
    def __init__(self, X, Y, model, modelParams = None, testX = None, testY = None,testTrainSplit = None,
                 folds = 5, hyperParams = None, scoring = None, performCV = None, targetEncodeCols = None):
        if (modelParams is None):
            modelParams = {}

        self.X = X
        self.Y = Y
        self.testX = testX
        self.testY = testY
        self.testTrainSplit = testTrainSplit
        self.folds = folds
        self.model = model
        self.hyperParams = hyperParams
        self.bestEstimator = None
        self.scoring = scoring
        self.targetEncodeCols = targetEncodeCols
        self.performCV = performCV
        self.modelParams = modelParams
        self.estimatorModel = self.getEstimatorModel()
        self.trainX, self.testX, self.trainY, self.testY = self.getTestTrainSlipt()

    @abc.abstractmethod
    def getTestTrainSlipt(self):
        pass

    def fit(self):
        if(self.hyperParams):
            self.__trainWithGridCV()
        elif(self.performCV):
            self.__trainWithCV()
        else:
            self.__trainNoCV()

    def __trainNoCV(self):
        self.bestEstimator = self.estimatorModel
        self.bestEstimator.fit(self.trainX, self.trainY)


    def __trainWithCV(self):
        crossValScores = cross_val_score(self.estimatorModel, self.trainX, y=list(self.trainY), scoring=self.scoring, cv=self.folds)
        print("Plain Cross Validation Scores")
        print(crossValScores)
        self.bestEstimator = self.estimatorModel
        self.bestEstimator.fit(self.trainX, self.trainY)

    def __trainWithGridCV(self):
        gridSearch = GridSearchCV(self.estimatorModel, self.hyperParams, cv=self.folds, scoring=self.scoring)
        gridSearch.fit(self.trainX, self.trainY)
        print("Cross Validation Grid Search Scores")
        cvres = gridSearch.cv_results_
        for meanTestScore, params in zip(cvres["mean_test_score"], cvres["params"]):
            print(meanTestScore, params)
        self.bestEstimator = gridSearch.best_estimator_

    def test(self):
        preds = self.bestEstimator.predict(self.testX)
        return preds

    @abc.abstractmethod
    def predict(self):
        pass

    def newDataPredict(self, testData):
        preds = self.bestEstimator.predict(testData)
        return preds

    @abc.abstractmethod
    def getEstimatorModel(self):
        pass



import abc

class MultiModelsPredictor:

    @abc.abstractmethod
    def __init__(self, X, Y, testX = None, testY = None,testTrainSplit = None,
                 folds = 5, scoring = None, performCV = None, targetEncodeCols = None):

        self.X = X
        self.Y = Y
        self.testX = testX
        self.testY = testY
        self.testTrainSplit = testTrainSplit
        self.folds = folds
        self.scoring = scoring
        self.targetEncodeCols = targetEncodeCols
        self.performCV = performCV

    @abc.abstractmethod
    def predict(self):
        pass








import abc
from onepiecepredictor.OnePieceRegression import OnePieceRegression
from onepiecepredictor.MultiModelsPredictor import MultiModelsPredictor


class MultiModelsRegression(MultiModelsPredictor):
    """
        This class can be used for hyper parameter tuning with cross validation and stratified splitting of data if required.

        X -> array-like(supported by Sklearn). If testTrainSplit is passed, this will be split into train and test
        Y -> array-like(supported by Sklearn). If testTrainSplit is passed, this will be split into train and test
        testX -> array-like(supported by Sklearn), test data. Ignored if testTrainSplit is passed
        testY -> array-like(supported by Sklearn), test data. Ignored if testTrainSplit is passed
        testTrainSplit -> float, ratio passed will be the amount of test data.
        performCV -> bool, Used when hyperParams not passed to perform plain CV.
        folds -> int, No of folds to be used for CV.
        scoring -> str, Evaluation metric. Currently supported values: r2,neg_mean_squared_error. If not passed r2 is used.
        targetEncodeCols -> List. List of columns to target encode.

    """

    def __init__(self, X, Y, testX = None, testY = None,testTrainSplit = None,
                 folds = 5, scoring = None, performCV = None, targetEncodeCols = None):
        super().__init__(X=X, Y=Y, testX=testX, testY=testY, testTrainSplit=testTrainSplit,
                         folds=folds, scoring=scoring, performCV=performCV, targetEncodeCols=targetEncodeCols)

    """
    Returns dictionary with keys as Models and Values as metric scores.
    """
    def predict(self):
        dummyRef = OnePieceRegression(X=self.X, Y=self.Y, model="LINEAR", modelParams=None,
                                      testTrainSplit=self.testTrainSplit,
                                      testX=self.testX, testY=self.testY, folds=self.folds, scoring=self.scoring,
                                      performCV=self.performCV, targetEncodeCols= self.targetEncodeCols)

        tempX = dummyRef.trainX
        tempY = dummyRef.trainY
        tempTestX = dummyRef.testX
        tempTestY = dummyRef.testY

        regressors = ["LINEAR","RF","SVM","KNN","ADABOOST","XGBOOST","CATBOOST"]
        resultsDict = {}
        for regressor in regressors:
            op = OnePieceRegression(X = tempX, Y = tempY, model = regressor, modelParams = {},testX = tempTestX, testY = tempTestY, testTrainSplit = None, folds = self.folds,
                                    scoring = self.scoring, performCV = self.performCV, targetEncodeCols = None)

            op.fit()
            score, preds = op.predict()
            resultsDict[regressor] = score

            del op

        return resultsDict







from setuptools import setup, find_packages

with open('README.md') as readme_file:
    README = readme_file.read()

with open('HISTORY.md') as history_file:
    HISTORY = history_file.read()

setup_args = dict(
    name='onepiecepredictor',
    version='1.1',
    description='Hyper Paramter Tuning and Models performance comparison',
    long_description_content_type="text/markdown",
    long_description=README + '\n\n' + HISTORY,
    license='MIT',
    packages=find_packages(),
    author='Vineel Kurma',
    author_email='vineel.prince7@gmail.com',
    keywords=['onepiecepredictor','hyper parameter tuning'],
    url='https://github.com/vkurma/onepiecepredictor',
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)

install_requires = [
    'numpy',
    'pandas',
    'scikit-learn',
    'xgboost',
    'catboost',
    'category_encoders',
    'imblearn',
]

if __name__ == '__main__':
    setup(**setup_args, install_requires=install_requires)


import abc
from onepiecepredictor.MultiModelsPredictor import MultiModelsPredictor
from onepiecepredictor.OnePieceClassifier import OnePieceClassifier

class MultiModelsClassifier(MultiModelsPredictor):
    """
        For Comparing multiple classification models performance with cross validation and stratified splitting of data if required.

        X -> array-like(supported by Sklearn). If testTrainSplit is passed, this will be split into train and test
        Y -> array-like(supported by Sklearn). If testTrainSplit is passed, this will be split into train and test
        testX -> array-like(supported by Sklearn), test data. Ignored if testTrainSplit is passed
        testY -> array-like(supported by Sklearn), test data. Ignored if testTrainSplit is passed
        testTrainSplit -> float, ratio passed will be the amount of test data.
        stratify -> bool, used to perform stratified splitting. If passed data will be split based on Y.
        performCV -> bool, Used when hyperParams not passed to perform plain CV.
        folds -> int, No of folds to be used for CV.
        applySmote -> bool, To apply smote to oversample the data. Pass only one of applySmote or underSample
        underSample -> bool, To randomly undersample the majority data.
        sampling -> str, Values supported by SMOTE, RandomUnderSampler classes in imblearn library.
        scoring -> str, Evaluation metric. Currently supported values: accuracy,balanced_accuracy,f1,precision,recall,roc_auc. If not passed accuracy is used.
        targetEncodeCols -> List. List of columns to target encode.
        multiClass -> Pass true in case of multiclass classification.
    """

    def __init__(self, X, Y, testX = None, testY = None,testTrainSplit = None,
                 folds = 5, scoring = None, performCV = None, targetEncodeCols = None,
                 applySmote=False, underSample=False, sampling=None, stratify=None, multiClass = False
                 ):
        self.multiClass = multiClass
        self.applySmote = applySmote
        self.sampling = sampling
        self.stratify = stratify
        self.underSample = underSample
        super().__init__(X=X, Y=Y, testX=testX, testY=testY, testTrainSplit=testTrainSplit,
                         folds=folds, scoring=scoring, performCV=performCV, targetEncodeCols=targetEncodeCols
                         )

    def predict(self):
        """
        Returns dictionary with keys as Models and Values as metric scores.
        """
        dummyRef = OnePieceClassifier(X = self.X,  Y = self.Y, model = "LOGISTIC", modelParams = None,testTrainSplit = self.testTrainSplit,
                                    testX = self.testX, testY = self.testY,folds = self.folds, scoring = self.scoring, performCV = self.performCV,
                                    targetEncodeCols = self.targetEncodeCols, applySmote = self.applySmote, underSample = self.underSample,
                                    sampling = self.sampling, stratify = self.stratify, multiClass = self.multiClass)

        tempX = dummyRef.trainX
        tempY = dummyRef.trainY
        tempTestX = dummyRef.testX
        tempTestY = dummyRef.testY

        classifiers = ["LOGISTIC","RF","SVM","KNN","ADABOOST","XGBOOST","CATBOOST"]
        resultsDict = {}
        for classifier in classifiers:
            op = OnePieceClassifier(X = tempX,  Y = tempY, model = classifier, modelParams = None, testTrainSplit = None,
                                    testX = tempTestX, testY = tempTestY, folds = self.folds, scoring = self.scoring, performCV = self.performCV,
                                    targetEncodeCols = None, applySmote = False, underSample = False,
                                    sampling = None, stratify = False, multiClass = self.multiClass)

            op.fit()
            score, preds = op.predict()
            resultsDict[classifier] = score

            del op

        return resultsDict



